#!/usr/bin/env python3

import argparse
import subprocess
import os
import sys
from yaml import dump
#import pandas as pd

version = "1.0.1"


def get_gitversion(fixed_version: str):
    cwd = os.getcwd()
    os.chdir(os.path.dirname(os.path.realpath(__file__)))
    try:
        git_hash = subprocess.check_output(["git", "describe", "--always"]).decode('ascii').strip()
    except FileNotFoundError:
        git_hash = fixed_version
    os.chdir(cwd)  # NOTE: logdir argument removed
    return git_hash


def check_sample_list(sample_list):
    # Check whether sample list has header
    type2 = "sample\tlong\tshort1\tshort2\n"

    with open (sample_list,'r') as handle:
        header_line = handle.readline()

    if header_line != type2 :
        print ("ERROR: sample list " + sample_list + " does not have a proper header. Make sure that first line is \nsample\tlong\tshort1\tshort2\n(tab-separated)")
        sys.exit(1)



def create_config(configfile, args, db_dir,repo_path):

    if not os.path.exists(args.sample_list):
        print("path to sample list " + args.sample_list + " does not exist")
        sys.exit(1)

    if not os.path.exists(args.kraken2db):
        print("path to kraken2 database " + args.kraken2db + " does not exist")
        sys.exit(1)

    if not os.path.exists(args.taxonkit_db):
        print("path to taxonkit_db database " + args.taxonkit_db + " does not exist")
        sys.exit(1)


    # check if sample list has proper format
    check_sample_list(args.sample_list)

# TODO check database paths
    #if not os.path.exists(args.mash_plasmid_db):
        #print("Path" + args.mash_plasmid_db + "does not exist")
        #sys.exit(1)

    #if not os.path.exists(args.mash_genome_db):
        #print("Path" + args.mash_genome_db + "does not exist")
        #sys.exit(1)

    #if not os.path.exists(args.mash_genome_info):
        #print("Path" + args.mash_genome_info + "does not exist")
        #sys.exit(1)

    if not os.path.exists(args.platon_db):
        print("Path to platon db " + args.platon_db + "does not exist")
        sys.exit(1)


    #print(args)
    # define do_pilon = True
    if args.no_pilon:
        do_pilon = False
    else:
        do_pilon = True


    if args.no_long and not args.no_flye:
        print("ERROR: You chose --no-long but not --no-flye")
        sys.exit(1)

    # define ASSEMBLER
    if args.no_unicycler and not args.no_flye:
        print("Only flye")
        ASSEMBLER = ["flye"]
    elif not args.no_unicycler and args.no_flye:
        print("Only unicycler")
        ASSEMBLER = ["unicycler"]
        # implies no pilon
        do_pilon = False
    elif not args.no_unicycler and not args.no_flye:
        print("flye and unicycler")
        ASSEMBLER = ["unicycler","flye"]
    else:
        print("Choose at least one assembler")
        sys.exit(1)
        
    # define READTYPE
    if args.no_short:
        print("Not using short reads")
        READTYPE = ["longreads"]
        # implies no pilon
        do_pilon = False
    elif args.no_long:
        READTYPE = ["shortreads"]
    else:
        READTYPE = ["shortreads","longreads"]
        do_pilon = not args.no_pilon


    if args.no_flye:
        do_pilon = False

    # define APPLIED_DATABASES
    APPLIED_DATABASES = args.abricate_databases # check existence

    if args.pacbio:
        technology = "PacBio"
        if args.flye_read_type == "--nano-raw":
            print('When using flag "--pacbio", please also change "flye_read-type" to \'--pacbio-raw\'   \'--pacbio-corr\' \'--pacbio-hifi\' ')
            sys.exit(1)
    elif args.no_long:
        print("Only performing short read assembly!")
        technology = "Illumina"
    else:
        technology = "ONT"


    #outfile = open(configfile, "w")
    #outfile.write("#Created by milonga.py v{}\n".format(version))
    
#    outfile.write("#loreapi git version is {}\n".format(git_version))


    #git_version = get_gitversion(args)
    git_version = get_gitversion(fixed_version=version)
    print("You are using milonga version " + git_version)


    config_dict = {
        'Created by MiLongA': version,
        'version'           : git_version,
        'workdir'           : args.working_directory,
        'samples'           : args.sample_list,
        'assembler'         : ASSEMBLER,
        'params'            : {
            'threads'    : args.threads_sample,
            'readtype'   : READTYPE,
            'technology' : technology,
            'trimming'   : {
                            'do_trimming': not args.no_trimming,
                            'minscore' : args.trimming_minscore},
            'filtering'  : {
                            'do_filtering': not args.no_filtering,
                            'minquality' : args.filtering_minquality,
                            'minlength' : args.filtering_minlength,
                            'extra' : args.filtering_extra },
            'unicycler'  : {
                            'do_unicycler' : not args.no_unicycler,
                            'mode' : args.unicycler_mode,
                            'keep' : args.unicycler_keep,
                            'extra' : args.unicycler_extra },
            'flye'        : {
                            'do_flye': not args.no_flye, 
                            'read_type' : args.flye_read_type, 
                            'iterations' : args.flye_iterations, 
                            'extra' : args.flye_extra },
            'abricate'    : {
                            'applied_databases' : APPLIED_DATABASES,
                            'path' : args.abricate_db,
                            'minid' : args.abricate_minid,
                            'mincov' : args.abricate_mincov,
                            'mincov_contigends' : args.abricate_mincov_contigends },
            'platon'      : {
                            'db' : args.platon_db},
            'pilon'       : {  
                            'do_pilon' : do_pilon,
                            'iterations' : args.pilon_iterations, 
                            'fix' : args.pilon_fix, 
                            'extra' : args.pilon_extra},
            'kraken2'     : { 'db_kraken': args.kraken2db,
                              'taxonkit_db': args.taxonkit_db,
                              'taxonkit_taxlevel': args.taxlevel,
                               'do_kraken': True }#,



        }
    }


    with open(configfile, 'w') as outfile:
        dump(config_dict, outfile, default_flow_style=False, sort_keys=False)



    ## OLD >>>>>>>>>>>>>>>>
    ## main specifications
    #outfile.write('workdir:\n  "{}"\n'.format(args.working_directory))
    #outfile.write('samples:\n  "{}"\n'.format(args.sample_list))
    #outfile.write('assembler:\n  {assembler}\n'.format(assembler = ASSEMBLER))

    ## general options
    #outfile.write('params:\n  threads: {threads}\n  readtype: {readtype}\n  technology: \"{technology}\"\n'.format(threads=args.threads_sample, readtype=READTYPE, technology = technology))
    ## trimming
    #outfile.write('  trimming:\n    do_trimming: {do_trimming}\n    minscore: {minscore}\n'.format(do_trimming=not args.no_trimming, minscore=args.trimming_minscore))
    ## filtering
    #outfile.write('  filtering:\n    do_filtering: {do_filtering}\n    minquality: {minquality}\n    minlength: {minlength}\n    extra: {extra}\n'.format(do_filtering=not args.no_filtering, minquality=args.filtering_minquality, minlength=args.filtering_minlength, extra=args.filtering_extra))
    ## unicycler
    #outfile.write('  unicycler:\n    do_unicycler: {do_unicycler}\n    mode: {mode}\n    keep: {keep}\n    extra: {extra}\n'.format(do_unicycler=not args.no_unicycler, mode=args.unicycler_mode, keep=args.unicycler_keep, extra=args.unicycler_extra))
    ## flye
    #outfile.write('  flye:\n    do_flye: {do_flye}\n    read_type: {read_type}\n    iterations: {iterations}\n    extra: {extra}\n'.format(do_flye=not args.no_flye, read_type=args.flye_read_type, iterations=args.flye_iterations, extra=args.flye_extra))
    ## abricate
    #outfile.write('  abricate:\n    applied_databases: {applied_databases}\n    path: {path}\n    minid: {minid}\n    mincov: {mincov}\n    mincov_contigends: {mincov_contigends}\n'.format(applied_databases = APPLIED_DATABASES, path = args.abricate_db, minid = args.abricate_minid, mincov = args.abricate_mincov, mincov_contigends = args.abricate_mincov_contigends))
    ## platon
    #outfile.write('  platon:\n    db: {}\n'.format(args.platon_db))
    ##outfile.write('  platon:\n    db: {}\n    mode: {}\n    options: {}\n'.format(args.platon_db, args.platon_mode,"--verbose"))
    ## pilon
    #outfile.write('  pilon:\n    do_pilon: {do_pilon}\n    iterations: {iterations}\n    fix: {fix}\n    extra: {extra}\n'.format(do_pilon = do_pilon, iterations = args.pilon_iterations, fix = args.pilon_fix, extra = args.pilon_extra))
    ## <<<<<<<<<<<<<<<<<<

    
    #outfile.close()


def run_snakemake(configfile, args):
    
    if args.forceall:
        forceall = "--forceall"
    else:
        forceall = ""

    if args.unlock:
        unlock = "--unlock"    
    else:
        unlock = ""   
    
    if args.dryrun:
        dryrun = "-n"
    else:
        dryrun = ""

    call = "snakemake --keep-going -p --configfile {configfile} --snakefile {snakefile} {forceall} {unlock} {dryrun} --cores {threads}".format(snakefile=args.snakefile, configfile=configfile, forceall=forceall, unlock = unlock, dryrun=dryrun, threads=args.threads)
    print(call)
    subprocess.call(call, shell=True)



def main():
    
   
    #find out path of abricate
    conda_path = os.environ['CONDA_PREFIX']
    db_dir = os.path.join(conda_path,"db")
    
    # find out repo path
    repo_path = os.path.dirname(os.path.realpath(__file__))

    
    parser = argparse.ArgumentParser()
    parser.add_argument('-l', '--sample_list', help='List of samples to analyze, as a two column tsv file with columns sample and assembly. Can be generated with provided script create_sampleSheet,sh', required=False, type=os.path.abspath)
    parser.add_argument('-d', '--working_directory', help='Working directory where results are saved', required=False, type=os.path.abspath)
    parser.add_argument('-c', '--config', help='Path to pre-defined config file (Other parameters will be ignored)', default="", required=False)

    parser.add_argument('-s', '--snakefile', help='Path to Snakefile of long read assembly pipeline, default is path to Snakefile in same directory', default = os.path.join(os.path.dirname(os.path.realpath(__file__)),"pipeline.smk"),required=False)

    # general parameters
    parser.add_argument('--no-trimming', help='Do not trim long reads', default=False, action='store_true',  required=False)
    parser.add_argument('--no-filtering', help='Do not perform quality and length filtering on long reads', default=False, action='store_true',  required=False)
    parser.add_argument('--no-unicycler', help='Do not Perform unicycler assembly', default=False, action='store_true',  required=False)
    parser.add_argument('--no-flye', help='Do not Perform flye assembly', default=False, action='store_true',  required=False)
    parser.add_argument('--no-pilon', help='Do not Perform pilon polising on flye assembly', default=False, action='store_true',  required=False)
    parser.add_argument('--no-short', help='Do not Perform mapping with short reads', default=False, action='store_true',  required=False)
    parser.add_argument('--no-long', help='Do only perorm unicycler assembly, with short reads only', default=False, action='store_true',  required=False)
    
    parser.add_argument('--pacbio', help='Long reads are pacbio reads. Do not run Nanostat', default=False, action='store_true',  required=False)



    # trimming
    parser.add_argument('--trimming_minscore', help='Minimum score to keep reads in trimming, default=60', default=60, required=False)
  # filtering:
    do_filtering: true
    parser.add_argument('--filtering_minquality', help='Minimum quality score of reads to keep reads in filtering, default=7', default=7, required=False)
    parser.add_argument('--filtering_minlength', help='Minimum length to keep reads in trimming, default=1000', default=1000, required=False)
    parser.add_argument('--filtering_extra', help='Extra parameters for filtering, default=""', default="", required=False)
  #unicycler:
    parser.add_argument('--unicycler_mode', help='unicycler mode option: Choose from conservative,normal,bold, default="normal"', default="normal", required=False)
    parser.add_argument('--unicycler_keep', help='unicycler keep option:  choose 0 or 1, see unicycler --help for more info, default=0', default=0, required=False)
    parser.add_argument('--unicycler_extra', help='Extra parameters for unicycler, enclose in \', default=""', default="", required=False)
  #flye:
    parser.add_argument('--flye_read_type', help='flye read type: choose from \'--pacbio-raw\'   \'--pacbio-corr\'   \'--pacbio-hifi\'   \'--nano-raw\'   \'--nano-corr\'   \'--subassemblies\', default="--nano-raw"', default="--nano-raw", required=False)
    parser.add_argument('--flye_iterations', help='Number of self-polising iterations in flye, default=3', default=3, required=False)
    parser.add_argument('--flye_extra', help='Extra parameters for flye, enclose in \', default=""', default="", required=False)
  #pilon:
    parser.add_argument('--pilon_fix', help='Type of errors to fix with pilon: choose from snps,indels,gaps,local,all,bases,none, default="all"', default="all", required=False)
    parser.add_argument('--pilon_iterations', help='Number of polishing iterations in pilon, default=5', default=5, required=False)
    parser.add_argument('--pilon_extra', help='Extra parameters for pilon, enclose in \', default=""', default="", required=False)
    # abricate parameters
    parser.add_argument('--abricate_databases', help='List of databases to search for genes, default=["ncbi","plasmidfinder"]', default=["ncbi","plasmidfinder"], required=False)
    parser.add_argument('--abricate_minid', help='Minimum identity in abricate, default=80', default=80, required=False)
    parser.add_argument('--abricate_mincov', help='Minimum coverage in abricate, default=80', default=80, required=False)
    parser.add_argument('--abricate_mincov_contigends', help='Minimum coverage in abricate, default=20', default=80, required=False)
    parser.add_argument('--abricate_db', help='Path to abricate and prokka databases, default is database of conda environment', default=db_dir,required=False, type=os.path.abspath)
# platon
    parser.add_argument('--platon_db', help='Path to location of platon database for plasmid prediction: plasmid_db; default: ' + os.path.join(os.path.dirname(os.path.realpath(__file__)), "databases/platon"), default=os.path.join(os.path.dirname(os.path.realpath(__file__)), "databases/platon"),required=False, type=os.path.abspath)
# extra
    parser.add_argument('--kraken2db', help='Path to kraken2 database; default: ' + os.path.join(os.path.dirname(os.path.realpath(__file__)), "databases", "kraken2"),
                        default=os.path.join(os.path.dirname(os.path.realpath(__file__)), "databases", "kraken2"), type=os.path.abspath, required=False)  # Note: minikraken2.tar.gz extracts to "../kraken/" not kraken2
    parser.add_argument('--taxonkit_db', help='Path to taxonkit_db; default:' + os.path.join(os.path.dirname(os.path.realpath(__file__)), "databases", "taxonkit"),
                        default=os.path.join(os.path.dirname(os.path.realpath(__file__)), "databases", "taxonkit"), type=os.path.abspath, required=False)
    parser.add_argument('--taxlevel', help='taxonomic level for kraken2 contig parsing, default="species"', default="species", required=False)





    # mash databases
    #parser.add_argument('--mash_genome_db', help='Path to mash db (default: path/2/repo/databases/mash/refseq.genomes.k21s1000.msh)', default=os.path.join(os.path.dirname(os.path.realpath(__file__)), "databases/mash/refseq.genomes.k21s1000.msh"),required=False, type=os.path.abspath)
    #parser.add_argument('--mash_genome_info', help='Path to mash db (default: path/2/repo/databases/mash/refseq.genomes.k21s1000.info)', default=os.path.join(os.path.dirname(os.path.realpath(__file__)), "databases/mash/refseq.genomes.k21s1000.info"),required=False, type=os.path.abspath)
    #parser.add_argument('--mash_plasmid_db', help='Path to mash db (default: path/2/repo/databases/mash/refseq.plasmid.k21s1000.msh)', default=os.path.join(os.path.dirname(os.path.realpath(__file__)), "databases/mash/refseq.plasmid.k21s1000.msh"),required=False, type=os.path.abspath)    
    # prokka parameters
    #parser.add_argument('--prokka', help='Also run prokka', default=False, action='store_true',  required=False)
    #parser.add_argument('--prokka_genus', help='Prokka genus option. See prokka --help for details. Default = "Genus"', default="Genus", required=False)

    parser.add_argument('-t', '--threads', help='Number of Threads to use. Note that samples can only be processed sequentially due to the required database access. However the allele calling can be executed in parallel for the different loci, default = 10', default=10, required=False)
    parser.add_argument('-n', '--dryrun', help='Snakemake dryrun. Only calculate graph without executing anything', default=False, action='store_true',  required=False)
    parser.add_argument('--threads_sample', help='Number of Threads to use per sample, default = 10', default=10, required=False, type=int)
    parser.add_argument('--forceall', help='Snakemake force. Force recalculation of all steps', default=False, action='store_true',  required=False)
    parser.add_argument('--unlock', help='unlock snakemake', default=False, action='store_true',  required=False)
    parser.add_argument('--logdir', help='Path to directory whete .snakemake output is to be saved', default='NA', required=False)

    args = parser.parse_args()


# write config if not exists
    if os.path.exists(args.config): # use pre-defined config
        configfile = args.config
        print("Using pre-defined configfile: " + configfile)
    else: # create new config
        configfile = os.path.join(args.working_directory, "config.yaml")
        if not os.path.exists(os.path.dirname(args.working_directory)):
        #if args.working_directory == "":
            print("Must specify -d / --working_directory path/2/workingdir")
            print("args.working_directory: " + args.working_directory + " upper directory expected to exist")
            sys.exit(1)
        if not os.path.isfile(args.sample_list):
            print("Must specify -l / --sample_list path/2/sample_list.tsv")
            sys.exit(1)
        if not os.path.exists(args.working_directory):
            os.makedirs(args.working_directory)
            # create config
        print("Creating new config: " + configfile)
        create_config(configfile,args,db_dir,repo_path)


    if not os.path.exists(args.snakefile):
        raise Exception('Snakefile not found')


    # run snakemake
    if os.path.exists(configfile):
        run_snakemake(configfile, args)
    else:
        print("Path to configfile does not exist: " + configfile)




if __name__ == '__main__':
    main()
